from .base import BaseFlow
from ..clients.asyncio import AsyncClient, AsyncWebSocketClient
from ..clients.mixin import AsyncPostMixin


class AsyncFlow(AsyncPostMixin, BaseFlow):
    """
    :class:`AsyncFlow` is the asynchronous version of the :class:`Flow`. They share the same interface, except
    in :class:`AsyncFlow` :meth:`train`, :meth:`index`, :meth:`search` methods are coroutines
    (i.e. declared with the async/await syntax), simply calling them will not schedule them to be executed.
    To actually run a coroutine, user need to put them in an eventloop, e.g. via ``asyncio.run()``,
    ``asyncio.create_task()``.

    :class:`AsyncFlow` can be very useful in
    the integration settings, where Jina/Jina Flow is NOT the main logic, but rather served as a part of other program.
    In this case, users often do not want to let Jina control the ``asyncio.eventloop``. On contrary, :class:`Flow`
    is controlling and wrapping the eventloop internally, making the Flow looks synchronous from outside.

    In particular, :class:`AsyncFlow` makes Jina usage in Jupyter Notebook more natural and reliable.
    For example, the following code
    will use the eventloop that already spawned in Jupyter/ipython to run Jina Flow (instead of creating a new one).

    .. highlight:: python
    .. code-block:: python

        from jina import AsyncFlow
        import numpy as np

        with AsyncFlow().add() as f:
            await f.index_ndarray(np.random.random([5, 4]), on_done=print)

    Notice that the above code will NOT work in standard Python REPL, as only Jupyter/ipython implements "autoawait".

    .. seealso::
        Asynchronous in REPL: Autoawait

        https://ipython.readthedocs.io/en/stable/interactive/autoawait.html

    Another example is when using Jina as an integration. Say you have another IO-bounded job ``heavylifting()``, you
    can use this feature to schedule Jina ``index()`` and ``heavylifting()`` concurrently.

    One can think of :class:`Flow` as Jina-managed eventloop, whereas :class:`AsyncFlow` is self-managed eventloop.
    """

    _cls_client = AsyncClient  #: the type of the Client, can be changed to other class

    def _update_client(self):
        if self._pod_nodes['gateway'].args.restful:
            self._cls_client = AsyncWebSocketClient



from typing import Optional, Iterator, Tuple, Dict, Iterable

import numpy as np

from . import Document, DocumentSourceType
from ..arrays import ChunkArray
from ..struct import StructView
from ..ndarray.sparse.scipy import SparseNdArray
from ...importer import ImportExtensions
from ...logging.predefined import default_logger

__all__ = ['GraphDocument']

if False:
    from scipy.sparse import coo_matrix
    from dgl import DGLGraph


class GraphDocument(Document):
    """
    :class:`GraphDocument` is a data type created based on Jina primitive data type :class:`Document`.

    It adds functionality that lets you work with a `Document` as a `directed graph` where all its chunks are the nodes in the `graph`.

    It exposes functionality to access and manipulate `graph related info` from the `DocumentProto` such as adjacency and edge features.

    .. warning::
        - It assumes that every ``chunk`` of a ``document`` is a node of a graph.

    :param document: the document to construct from. If ``bytes`` is given
            then deserialize a :class:`DocumentProto`; ``dict`` is given then
            parse a :class:`DocumentProto` from it; ``str`` is given, then consider
            it as a JSON string and parse a :class:`DocumentProto` from it; finally,
            one can also give `DocumentProto` directly, then depending on the ``copy``,
            it builds a view or a copy from it.
    :param copy: when ``document`` is given as a :class:`DocumentProto` object, build a
            view (i.e. weak reference) from it or a deep copy from it.
    :param kwargs: further key value arguments
    """

    def __init__(
        self,
        document: Optional[DocumentSourceType] = None,
        copy: bool = False,
        **kwargs,
    ):
        self._check_installed_array_packages()
        super().__init__(document=document, copy=copy, **kwargs)
        self._node_id_to_offset = {
            node.id: offset for offset, node in enumerate(self.nodes)
        }  # dangerous because document is stateless, try to work only with proto

    @staticmethod
    def _check_installed_array_packages():
        from ... import JINA_GLOBAL

        if JINA_GLOBAL.scipy_installed is None:
            JINA_GLOBAL.scipy_installed = False
            with ImportExtensions(
                required=True,
                pkg_name='scipy',
                help_text=f'GraphDocument requires scipy to be installed for sparse matrix support.',
            ):
                import scipy

                JINA_GLOBAL.scipy_installed = True

    def add_node(self, node: 'Document'):
        """
        Add a a node to the graph

        :param node: the node to be added to the graph
        """
        if node.id in self._node_id_to_offset:
            default_logger.warning(f'Document {node.id} is already a node of the graph')
            return

        self._node_id_to_offset[node.id] = len(self.nodes)
        self.nodes.append(node)

    def remove_node(self, node: 'Document'):
        """
        Remove a node from the graph along with the edges that may contain it

        :param node: the node to be removed from the graph
        """
        from scipy.sparse import coo_matrix

        if node.id not in self._node_id_to_offset:
            default_logger.warning(
                f'Trying to remove document {node.id} from the graph while is not a node of the graph'
            )
            return

        offset = self._node_id_to_offset[node.id]

        if self.num_edges > 0:
            edges_to_remove = []
            for edge_id, (row, col) in enumerate(
                zip(self.adjacency.row, self.adjacency.col)
            ):
                if row.item() == offset or col.item() == offset:
                    edge_features_keys = (
                        f'{self.nodes[row.item()].id}-{self.nodes[col.item()]}'
                    )
                    edges_to_remove.append((edge_id, edge_features_keys))

            for edge_id, edge_features_key in reversed(edges_to_remove):
                self._remove_edge_id(edge_id, edge_features_key)

            if self.num_edges > 0:
                row = np.copy(self.adjacency.row)
                col = np.copy(self.adjacency.col)
                data = np.copy(self.adjacency.data)
                for i in range(self.num_edges):
                    if self.adjacency.row[i] > offset:
                        row[i] = row[i] - 1
                    if self.adjacency.col[i] > offset:
                        col[i] = col[i] - 1
                self.adjacency = coo_matrix((data, (row, col)))

        del self.nodes[offset]
        self._node_id_to_offset = {
            node.id: offset for offset, node in enumerate(self.nodes)
        }

    def add_edge(
        self, doc1: 'Document', doc2: 'Document', features: Optional[Dict] = None
    ):
        """
        Add an edge to the graph connecting `doc1` with `doc2`

        :param doc1: the starting node for this edge
        :param doc2: the ending node for this edge
        :param features: Optional features dictionary to be added to this new created edge
        """
        from scipy.sparse import coo_matrix

        self.add_node(doc1)
        self.add_node(doc2)

        current_adjacency = self.adjacency
        doc1_node_offset = self._node_id_to_offset[doc1.id]
        doc2_node_offset = self._node_id_to_offset[doc2.id]
        row = (
            np.append(current_adjacency.row, doc1_node_offset)
            if current_adjacency is not None
            else np.array([doc1_node_offset])
        )
        col = (
            np.append(current_adjacency.col, doc2_node_offset)
            if current_adjacency is not None
            else np.array([doc2_node_offset])
        )
        data = (
            np.append(current_adjacency.data, 1)
            if current_adjacency is not None
            else np.array([1])
        )
        self.adjacency = coo_matrix((data, (row, col)))
        if features is not None:
            self.edge_features[f'{doc1.id}-{doc2.id}'] = features

    def _remove_edge_id(self, edge_id: int, edge_feature_key: str):
        from scipy.sparse import coo_matrix

        if self.adjacency is not None:
            if edge_id > self.num_edges:
                raise Exception(
                    f'Trying to remove edge {edge_id} while number of edges is {self.num_edges}'
                )
            row = np.delete(self.adjacency.row, edge_id)
            col = np.delete(self.adjacency.col, edge_id)
            data = np.delete(self.adjacency.data, edge_id)
            if row.shape[0] > 0:
                self.adjacency = coo_matrix((data, (row, col)))
            else:
                self.adjacency = coo_matrix((0, 0))

            if edge_feature_key in self.edge_features:
                del self.edge_features[edge_feature_key]

    def remove_edge(self, doc1: 'Document', doc2: 'Document'):
        """
        Remove a node from the graph along with the edges that may contain it

        :param doc1: the starting node for this edge
        :param doc2: the ending node for this edge
        """
        offset1 = self._node_id_to_offset[doc1.id]
        offset2 = self._node_id_to_offset[doc2.id]
        for edge_id, (row, col) in enumerate(
            zip(self.adjacency.row, self.adjacency.col)
        ):
            if row.item() == offset1 and col.item() == offset2:
                self._remove_edge_id(edge_id, f'{doc1.id}-{doc2.id}')

    @property
    def edge_features(self):
        """
        The dictionary of edge features, indexed by `edge_id` in the `edge list`

        .. # noqa: DAR201
        """
        return StructView(self._pb_body.graph.edge_features)

    @edge_features.setter
    def edge_features(self, value: Dict):
        """Set the `edge_features` field of this Graph to a Python dict

        :param value: a Python dict
        """
        self._pb_body.graph.edge_features.Clear()
        self._pb_body.graph.edge_features.update(value)

    @property
    def adjacency(self):
        """
        The adjacency list for this graph,

        .. # noqa: DAR201
        """
        return SparseNdArray(self._pb_body.graph.adjacency, sp_format='coo').value

    @adjacency.setter
    def adjacency(self, value: 'coo_matrix'):
        """
        Set the adjacency list of this graph.

        :param value: the float weight of the document.
        """
        SparseNdArray(self._pb_body.graph.adjacency, sp_format='coo').value = value

    @property
    def num_nodes(self) -> int:
        """
        The number of nodes in the graph

        .. # noqa: DAR201
        """
        return len(self.nodes)

    @property
    def num_edges(self) -> int:
        """
        The number of edges in the graph

        .. # noqa: DAR201
        """
        adjacency = self.adjacency
        return adjacency.data.shape[0] if adjacency is not None else 0

    @property
    def nodes(self):
        """
        The nodes list for this graph

        .. # noqa: DAR201
        """
        return self.chunks

    def get_out_degree(self, doc: 'Document') -> int:
        """
        The out degree of the doc node

        .. # noqa: DAR201
        :param doc: the document node from which to extract the outdegree.
        """
        out_edges = self.get_outgoing_nodes(doc)
        return len(out_edges) if out_edges else 0

    def get_in_degree(self, doc: 'Document') -> int:
        """
        The in degree of the doc node

        .. # noqa: DAR201
        :param doc: the document node from which to extract the indegree.
        """
        in_edges = self.get_incoming_nodes(doc)
        return len(in_edges) if in_edges else 0

    @nodes.setter
    def nodes(self, value: Iterable['Document']):
        """Set all nodes of the current document.

        :param value: the array of nodes of this document
        """
        self.chunks = value

    @adjacency.setter
    def adjacency(self, value: "coo_matrix"):
        """
        Set the adjacency list of this graph.

        :param value: the float weight of the document.
        """
        SparseNdArray(self._pb_body.graph.adjacency, sp_format='coo').value = value

    def get_outgoing_nodes(self, doc: 'Document') -> Optional[ChunkArray]:
        """
        Get all the outgoing edges from `doc`

        .. # noqa: DAR201
        :param doc: the document node from which to extract the outgoing nodes.
        """
        if self.adjacency is not None and doc.id in self._node_id_to_offset:
            offset = self._node_id_to_offset[doc.id]
            return ChunkArray(
                [
                    self.nodes[col.item()]
                    for (row, col) in zip(self.adjacency.row, self.adjacency.col)
                    if row.item() == offset
                ],
                reference_doc=self,
            )

    def get_incoming_nodes(self, doc: 'Document') -> Optional[ChunkArray]:
        """
        Get all the outgoing edges from `doc`

        .. # noqa: DAR201
        :param doc: the document node from which to extract the incoming nodes.
        """
        if self.adjacency is not None and doc.id in self._node_id_to_offset:
            offset = self._node_id_to_offset[doc.id]
            return ChunkArray(
                [
                    self.nodes[row.item()]
                    for (row, col) in zip(self.adjacency.row, self.adjacency.col)
                    if col.item() == offset
                ],
                reference_doc=self,
            )

    @staticmethod
    def load_from_dgl_graph(dgl_graph: 'DGLGraph') -> 'GraphDocument':
        """
        Construct a GraphDocument from of graph with type `DGLGraph`

        .. # noqa: DAR201
        :param dgl_graph: the graph from which to construct a `GraphDocument`.

        .. warning::
        - This method only deals with the graph structure (nodes and conectivity) graph
          features that are task specific  are ignored.
        """
        jina_graph = GraphDocument()
        nodeid_to_doc = {}
        for node in dgl_graph.nodes():
            node_doc = Document()
            nodeid_to_doc[int(node)] = node_doc
            jina_graph.add_node(node_doc)

        for node_source, node_destination in zip(*dgl_graph.edges()):
            jina_graph.add_edge(
                nodeid_to_doc[int(node_source)], nodeid_to_doc[int(node_destination)]
            )

        return jina_graph

    def to_dgl_graph(self) -> 'DGLGraph':
        """
        Construct a  `dgl.DGLGraph` from a `GraphDocument` instance.

        .. warning::
        - This method only deals with the graph structure (nodes and conectivity) graph
          features that are task specific are ignored.

        .. # noqa: DAR201
        """
        from ... import JINA_GLOBAL

        if JINA_GLOBAL.dgl_installed is None:
            JINA_GLOBAL.dgl_installed = False
            with ImportExtensions(
                required=True,
                pkg_name='dgl',
                help_text=f'to_dgl_graph method requires dgl to be installed',
            ):
                import dgl

                JINA_GLOBAL.dgl_installed = True

        if JINA_GLOBAL.torch_installed is None:
            JINA_GLOBAL.torch_installed = False
            with ImportExtensions(
                required=True,
                pkg_name='torch',
                help_text=f'to_dgl_graph method requires torch to be installed',
            ):
                import torch

                JINA_GLOBAL.torch_installed = True

        import torch
        import dgl

        source_nodes = torch.tensor(self.adjacency.row.copy())
        destination_nodes = torch.tensor(self.adjacency.col.copy())

        return dgl.graph((source_nodes, destination_nodes))

    def __iter__(self) -> Iterator[Tuple['Document']]:
        for (row, col) in zip(self.adjacency.row, self.adjacency.col):
            yield self.nodes[row.item()], self.nodes[col.item()]

    def __mermaid_str__(self):
        results = []
        printed_ids = set()
        _node_id_node_mermaid_id = {}

        for node in self.nodes:
            _node_id_node_mermaid_id[node.id] = node._mermaid_id

        for in_node, out_node in self:

            in_node_mermaid_id = _node_id_node_mermaid_id[in_node.id]
            if in_node_mermaid_id not in printed_ids:
                in_node._mermaid_id = in_node_mermaid_id
                printed_ids.add(in_node_mermaid_id)
                results.append(in_node.__mermaid_str__())

            out_node_mermaid_id = _node_id_node_mermaid_id[out_node.id]
            if out_node_mermaid_id not in printed_ids:
                out_node._mermaid_id = out_node_mermaid_id
                printed_ids.add(out_node_mermaid_id)
                results.append(out_node.__mermaid_str__())

            results.append(f'{in_node_mermaid_id[:3]} --> {out_node_mermaid_id[:3]}')

        return '\n'.join(results)



"""
Top-level module of Jina.

The primary function of this module is to import all of the public Jina
interfaces into a single place. The interfaces themselves are located in
sub-modules, as described below.

"""

import datetime as _datetime
import os as _os
import platform as _platform
import signal as _signal
import sys as _sys
import types as _types

if _sys.version_info < (3, 7, 0) or _sys.version_info >= (3, 10, 0):
    raise OSError(f'Jina requires Python 3.7/3.8/3.9, but yours is {_sys.version_info}')

# DO SOME OS-WISE PATCHES
if _sys.version_info >= (3, 8, 0) and _platform.system() == 'Darwin':
    # temporary fix for python 3.8 on macos where the default start is set to "spawn"
    # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
    from multiprocessing import set_start_method as _set_start_method

    _set_start_method('fork')

# fix fork error on MacOS but seems no effect? must do EXPORT manually before jina start
_os.environ['OBJC_DISABLE_INITIALIZE_FORK_SAFETY'] = 'YES'

# do not change this line manually
# this is managed by git tag and updated on every release
# NOTE: this represents the NEXT release version

# TODO: remove 'rcN' on final release
__version__ = '2.0.0rc5.dev72'

# do not change this line manually
# this is managed by proto/build-proto.sh and updated on every execution
__proto_version__ = '0.0.81'

__uptime__ = _datetime.datetime.now().isoformat()

# update on MacOS
# 1. clean this tuple,
# 2. grep -rohEI --exclude-dir=jina/hub --exclude-dir=tests --include \*.py "\'JINA_.*?\'" jina  | sort -u | sed "s/$/,/g"
# 3. copy all lines EXCEPT the first (which is the grep command in the last line)
__jina_env__ = (
    'JINA_ARRAY_QUANT',
    'JINA_CONTROL_PORT',
    'JINA_DEFAULT_HOST',
    'JINA_DISABLE_UVLOOP',
    'JINA_EXECUTOR_WORKDIR',
    'JINA_FULL_CLI',
    'JINA_IPC_SOCK_TMP',
    'JINA_LOG_CONFIG',
    'JINA_LOG_ID',
    'JINA_LOG_LEVEL',
    'JINA_LOG_NO_COLOR',
    'JINA_LOG_WORKSPACE',
    'JINA_OPTIMIZER_TRIAL_WORKSPACE',
    'JINA_POD_NAME',
    'JINA_RANDOM_PORTS',
    'JINA_RANDOM_PORT_MAX',
    'JINA_RANDOM_PORT_MIN',
    'JINA_SOCKET_HWM',
    'JINA_VCS_VERSION',
    'JINA_WARN_UNNAMED',
)

__default_host__ = _os.environ.get('JINA_DEFAULT_HOST', '0.0.0.0')
__default_executor__ = 'BaseExecutor'
__default_endpoint__ = '/default'
__ready_msg__ = 'ready and listening'
__stop_msg__ = 'terminated'
__num_args_executor_func__ = 5
__root_dir__ = _os.path.dirname(_os.path.abspath(__file__))
__resources_path__ = _os.path.join(
    _os.path.dirname(_sys.modules['jina'].__file__), 'resources'
)


_names_with_underscore = [
    '__version__',
    '__proto_version__',
    '__default_host__',
    '__ready_msg__',
    '__stop_msg__',
    '__jina_env__',
    '__uptime__',
    '__root_dir__',
    '__default_endpoint__',
    '__default_executor__',
    '__num_args_executor_func__',
]


# ADD GLOBAL NAMESPACE VARIABLES
JINA_GLOBAL = _types.SimpleNamespace()
JINA_GLOBAL.scipy_installed = None
JINA_GLOBAL.tensorflow_installed = None
JINA_GLOBAL.torch_installed = None
JINA_GLOBAL.dgl_installed = None

# import jina.importer as _ji
#
# _ji.import_classes('jina.executors', show_import_table=False, import_once=True)
#
_signal.signal(_signal.SIGINT, _signal.default_int_handler)


def _set_nofile(nofile_atleast=4096):
    """
    Set nofile soft limit to at least 4096, useful for running matlplotlib/seaborn on
    parallel executing plot generators vs. Ubuntu default ulimit -n 1024 or OS X El Captian 256
    temporary setting extinguishing with Python session.

    :param nofile_atleast: nofile soft limit
    :return: nofile soft limit and nofile hard limit
    """

    try:
        import resource as res
    except ImportError:  # Windows
        res = None

    if res is None:
        return (None,) * 2

    soft, ohard = res.getrlimit(res.RLIMIT_NOFILE)
    hard = ohard

    if soft < nofile_atleast:
        soft = nofile_atleast
        if hard < soft:
            hard = soft

        try:
            res.setrlimit(res.RLIMIT_NOFILE, (soft, hard))
        except (ValueError, res.error):
            try:
                hard = soft
                print(f'trouble with max limit, retrying with soft,hard {soft},{hard}')
                res.setrlimit(res.RLIMIT_NOFILE, (soft, hard))
            except Exception:
                print('failed to set ulimit, giving up')
                soft, hard = res.getrlimit(res.RLIMIT_NOFILE)

    return soft, hard


_set_nofile()

# ONLY FIRST CLASS CITIZENS ARE ALLOWED HERE, namely Document, Executor Flow

# Client
from jina.clients import Client

# Document
from jina.types.document import Document
from jina.types.arrays.document import DocumentArray

# Executor
from jina.executors import BaseExecutor as Executor
from jina.executors.decorators import requests

# Flow
from jina.flow import Flow
from jina.flow.asyncio import AsyncFlow

__all__ = [_s for _s in dir() if not _s.startswith('_')]
__all__.extend(_names_with_underscore)



import os
from collections import defaultdict
from typing import Optional

import yaml

from .parameters import (
    IntegerParameter,
    UniformParameter,
    LogUniformParameter,
    CategoricalParameter,
    DiscreteUniformParameter,
)
from .parameters import load_optimization_parameters
from ..helper import colored
from ..importer import ImportExtensions
from ..jaml import JAMLCompatible, JAML
from ..logging.predefined import default_logger as logger

if False:
    from .flow_runner import FlowRunner
    import optuna
    from optuna.trial import Trial
    from argparse import Namespace


class OptimizerCallback(JAMLCompatible):
    """
    Callback interface for storing and calculating evaluation metric during an optimization.
    Should be used, whenever a custom evaluation aggregation during an Flow optimization is needed.
    """

    def get_empty_copy(self) -> 'OptimizerCallback':
        """
        Return an empty copy of the :class:`OptimizerCallback`.

        .. # noqa: DAR202
        :raises NotImplementedError: :class:`OptimizerCallback` is just an interface. Please use any implemented subclass.
        """
        raise NotImplementedError

    def get_final_evaluation(self) -> float:
        """
        Return the aggregation of all evaluation collected via :method:`__call__`

        .. # noqa: DAR202
        :raises NotImplementedError: :class:`OptimizerCallback` is just an interface. Please use any implemented subclass.
        """
        raise NotImplementedError

    def __call__(self, response):
        """
        Collects the results of evaluators in the response object for aggregation.

        :param response: A response object of a Flow.
        :raises NotImplementedError: :class:`OptimizerCallback` is just an interface. Please use any implemented subclass.
        """
        raise NotImplementedError


class MeanEvaluationCallback(OptimizerCallback):
    """
    Calculates the mean of all evaluations during a single :py:class:`FlowRunner`
    execution from the :py:class:`FlowOptimizer`.
    """

    def __init__(self, eval_name: Optional[str] = None):
        """
        :param eval_name: evaluation name as required by the evaluator. Not needed if only 1 evaluator is used
        """
        self._eval_name = eval_name
        self._evaluation_values = defaultdict(float)
        self._n_docs = 0

    def get_empty_copy(self):
        """
        Return an empty copy of the :class:`OptimizerCallback`.

        :return: Evaluation values
        """
        return MeanEvaluationCallback(self._eval_name)

    def get_final_evaluation(self):
        """
        Calculates and returns mean evaluation value on the metric defined in the :method:`__init__`.

        :return:: The aggregation of all evaluation collected via :method:`__call__`
        """
        if self._eval_name is not None:
            evaluation_name = self._eval_name
        else:
            evaluation_name = list(self._evaluation_values)[0]
            if len(self._evaluation_values) > 1:
                logger.warning(
                    f'More than one evaluation metric found. Please define the right eval_name. Currently {evaluation_name} is used'
                )

        return self._evaluation_values[evaluation_name] / self._n_docs

    def __call__(self, response):
        """
        Store the evaluation values

        :param response: response message
        """
        self._n_docs += len(response.data.docs)
        logger.info(f'Num of docs evaluated: {self._n_docs}')
        for doc in response.data.docs:
            for evaluation in doc.evaluations:
                self._evaluation_values[evaluation.op_name] += evaluation.value


class ResultProcessor(JAMLCompatible):
    """Result processor for the Optimizer."""

    def __init__(self, study: 'optuna.study.Study'):
        """
        :param study: optuna study object
        """
        self._study = study
        self._best_parameters = study.best_trial.params
        logger.info(colored(f'Number of finished trials: {len(study.trials)}', 'green'))
        logger.info(colored(f'Best trial: {study.best_trial.params}', 'green'))
        logger.info(colored(f'Time to finish: {study.best_trial.duration}', 'green'))

    @property
    def study(self):
        """
        :return:: Raw optuna study as calculated by the :py:class:`FlowOptimizer`.
        """
        return self._study

    @property
    def best_parameters(self):
        """
        :return:: The parameter set, which got the best evaluation result during the optimization.
        """
        return self._best_parameters

    def save_parameters(self, filepath: str = 'config/best_config.yml'):
        """
        Stores the best parameters in the given file.

        :param filepath: path where the best parameter config will be saved
        """
        parameter_dir = os.path.dirname(filepath)
        os.makedirs(parameter_dir, exist_ok=True)
        yaml.dump(self.best_parameters, open(filepath, 'w'))


class FlowOptimizer(JAMLCompatible):
    """
    Optimizer runs the given flows on multiple parameter configurations in order
    to find the best performing parameters. Uses `optuna` behind the scenes.
    For a detailed information how the parameters are sampled by optuna see
    https://optuna.readthedocs.io/en/stable/reference/generated/optuna.trial.Trial.html
    """

    def __init__(
        self,
        flow_runner: 'FlowRunner',
        parameter_yaml: str,
        evaluation_callback: 'OptimizerCallback',
        n_trials: int,
        workspace_base_dir: str = '.',
        sampler: str = 'TPESampler',
        direction: str = 'maximize',
        seed: int = 42,
    ):
        """
        :param flow_runner: `FlowRunner` object which contains the flows to be run.
        :param parameter_yaml: yaml container the parameters to be optimized
        :param evaluation_callback: The callback object, which stores the evaluation results
        :param n_trials: evaluation trials to be run
        :param workspace_base_dir: directory in which all temporary created data should be stored
        :param sampler: The optuna sampler. For a list of usable names see: https://optuna.readthedocs.io/en/stable/reference/samplers.html
        :param direction: direction of the optimization from either of `maximize` or `minimize`
        :param seed: random seed for reproducibility
        """
        super().__init__()
        self._version = '1'
        self._flow_runner = flow_runner
        self._parameter_yaml = parameter_yaml
        self._workspace_base_dir = workspace_base_dir
        self._evaluation_callback = evaluation_callback
        self._n_trials = n_trials
        self._sampler = sampler
        self._direction = direction
        self._seed = seed

    def _trial_parameter_sampler(self, trial: 'Trial'):
        trial_parameters = {}
        parameters = load_optimization_parameters(self._parameter_yaml)
        for param in parameters:
            trial_parameters[param.jaml_variable] = param.suggest(trial)

        trial.workspace = (
            self._workspace_base_dir
            + '/JINA_WORKSPACE_'
            + '_'.join([str(v) for v in trial_parameters.values()])
        )

        trial_parameters['JINA_OPTIMIZER_TRIAL_WORKSPACE'] = trial.workspace

        return trial_parameters

    def _objective(self, trial: 'Trial'):
        trial_parameters = self._trial_parameter_sampler(trial)
        evaluation_callback = self._evaluation_callback.get_empty_copy()
        self._flow_runner.run(
            trial_parameters, workspace=trial.workspace, callback=evaluation_callback
        )
        eval_score = evaluation_callback.get_final_evaluation()
        logger.info(colored(f'Evaluation Score: {eval_score}', 'green'))
        return eval_score

    def optimize_flow(self, **kwargs) -> 'ResultProcessor':
        """
        Will run the actual optimization.

        :param kwargs: extra parameters for optuna sampler
        :return:: The aggregated result of the optimization run as a :class:`ResultProcessor`.
        """
        with ImportExtensions(required=True):
            import optuna
        if self._sampler == 'GridSampler':
            sampler = getattr(optuna.samplers, self._sampler)(**kwargs)
        else:
            sampler = getattr(optuna.samplers, self._sampler)(seed=self._seed, **kwargs)
        study = optuna.create_study(direction=self._direction, sampler=sampler)
        study.optimize(self._objective, n_trials=self._n_trials)
        result_processor = ResultProcessor(study)
        return result_processor


def run_optimizer_cli(args: 'Namespace'):
    """
    Used to run the FlowOptimizer from command line interface.

    :param args: arguments passed via cli
    """
    # The following import is needed to initialize the JAML parser
    from .flow_runner import SingleFlowRunner, MultiFlowRunner

    with open(args.uses) as f:
        optimizer = JAML.load(f)
    result_processor = optimizer.optimize_flow()
    if args.output_dir:
        result_processor.save_parameters(
            os.path.join(args.output_dir, 'best_parameters.yml')
        )
